from __future__ import annotations

import asyncio
import typing

from ctc import rpc
from ctc import evm
from ctc import spec


guni_function_abis = {
    'token0': {
        'constant': True,
        'inputs': [],
        'name': 'token0',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'token1': {
        'constant': True,
        'inputs': [],
        'name': 'token1',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
}


async def async_get_tokens(
    g_uni_pool: spec.Address,
    *,
    context: spec.Context = None,
) -> tuple[spec.Address, spec.Address]:
    return await asyncio.gather(
        rpc.async_eth_call(
            to_address=g_uni_pool, function_name='token0', context=context
        ),
        rpc.async_eth_call(
            to_address=g_uni_pool, function_name='token1', context=context
        ),
    )


async def async_get_token_balances(
    g_uni_pool: spec.Address,
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Sequence[int | float]:
    balances_coroutine = rpc.async_eth_call(
        to_address=g_uni_pool,
        function_name='getUnderlyingBalances',
        context=context,
    )
    balances_task = asyncio.create_task(balances_coroutine)

    if normalize:
        tokens_coroutine = async_get_tokens(
            g_uni_pool,
            context=context,
        )
        tokens_task = asyncio.create_task(tokens_coroutine)

        return await evm.async_normalize_erc20s_quantities(
            quantities=(await balances_task),
            tokens=(await tokens_task),
            context=context,
        )

    else:

        result = await balances_task
        if not isinstance(result, (tuple, list)) or not all(
            isinstance(item, int) for item in result
        ):
            raise Exception('invalid rpc result')
        return result


async def async_get_token_balances_by_block(
    g_uni_pool: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Sequence[typing.Sequence[int | float]]:
    balances_coroutine = rpc.async_batch_eth_call(
        to_address=g_uni_pool,
        function_name='getUnderlyingBalances',
        block_numbers=blocks,
        context=context,
    )
    balances_task = asyncio.create_task(balances_coroutine)

    if normalize:
        tokens_coroutine = async_get_tokens(g_uni_pool, context=context)
        tokens_task = asyncio.create_task(tokens_coroutine)

        balances = await balances_task
        token0_balances, token1_balances = list(zip(*balances))

        # normalize
        tokens = await tokens_task
        token0_coroutine = evm.async_normalize_erc20_quantities_by_block(
            quantities=token0_balances,
            token=tokens[0],
            blocks=blocks,
            context=context,
        )
        token1_coroutine = evm.async_normalize_erc20_quantities_by_block(
            quantities=token1_balances,
            token=tokens[1],
            blocks=blocks,
            context=context,
        )
        token0_task = asyncio.create_task(token0_coroutine)
        token1_task = asyncio.create_task(token1_coroutine)

        result = await asyncio.gather(token0_task, token1_task)
        return list(list(item) for item in zip(*result))

    else:
        return await balances_task




"""
range bounds are **inclusive** by default
- can be switched to python-index-style ranges using index=True
"""
from __future__ import annotations

import math
import typing

if typing.TYPE_CHECKING:
    Range = typing.Sequence[int]

    T = typing.TypeVar('T')


def split(
    items: typing.Sequence[T],
    n_splits: int | None = None,
    *,
    items_per_split: int | None = None,
    exact: bool = False,
) -> typing.Sequence[typing.Sequence[T]]:

    if n_splits is not None and items_per_split is not None:
        raise Exception('specify either n_splits or items_per_split')

    elif n_splits is not None:

        n_items = len(items)
        items_per_split = math.floor(n_items / n_splits)

        if exact and n_items != n_splits * items_per_split:
            raise Exception('n_splits does not exactly divide items')

        sizes = [items_per_split] * n_splits

        remaining = n_items - items_per_split * n_splits
        for r in range(remaining):
            sizes[r] += 1

        splits = []
        current = 0
        for size in sizes:
            next = current + size
            sl = slice(current, next)
            splits.append(items[sl])
            current = next

        return splits

    elif items_per_split is not None:
        n_items = len(items)
        n_splits = math.ceil(n_items / items_per_split)
        if exact and n_items != n_splits * items_per_split:
            raise Exception('n_splits does not exactly divide items')
        splits = []
        for s in range(n_splits):
            sl = slice(s * items_per_split, (s + 1) * items_per_split)
            splits.append(items[sl])
        return splits

    else:
        raise Exception('specify either n_splits or items_per_split')


def get_range_gaps(
    *,
    start: int,
    end: int,
    subranges: typing.Sequence[Range],
) -> typing.Sequence[Range]:
    """get gaps between subranges of a given range"""

    if len(subranges) == 0:
        return [[start, end]]

    # make ranges non-overlapping
    subranges = combine_overlapping_ranges(subranges)

    # sort subranges
    subranges = sorted(subranges, key=lambda subrange: subrange[0])

    # get gap ranges
    gap_ranges = []
    current_start = start
    for substart, subend in subranges:

        # skip subranges that fall out of main range
        if subend < current_start:
            continue
        if substart > end:
            break

        if substart <= current_start and subend >= current_start:
            current_start = subend + 1
            continue
        else:
            gap_range = [current_start, substart - 1]
            gap_ranges.append(gap_range)
            current_start = subend + 1
    if current_start < end:
        gap_ranges.append([current_start, end])

    if (
        len(gap_ranges) > 0
        and gap_ranges[-1][-1] < end
        and subranges[-1][-1] < end
    ):
        gap_range = [subend + 1, end]
        gap_ranges.append(gap_range)

    return gap_ranges


def get_overlapping_ranges(
    ranges: typing.Sequence[Range],
    *,
    include_contiguous: bool = False,
) -> typing.Sequence[tuple[int, int]]:
    """get pairs of ranges that have overlap

    if include_contiguous, then consider touching ranges to be overlapping
    """

    overlapping_ranges = []
    for i, (start, end) in enumerate(ranges):
        other_slice = i + 1
        for j, (other_start, other_end) in list(enumerate(ranges))[
            other_slice:
        ]:
            if start <= other_end and other_start <= end:
                overlapping_ranges.append((i, j))
            elif include_contiguous and (
                start == other_end + 1 or other_start == end + 1
            ):
                overlapping_ranges.append((i, j))
    return overlapping_ranges


def combine_overlapping_ranges(
    ranges: typing.Sequence[Range],
    *,
    include_contiguous: bool = False,
) -> typing.Sequence[Range]:
    """combine ranges as needed to produce list of non-overlapping ranges"""

    overlapping_ranges = get_overlapping_ranges(
        ranges,
        include_contiguous=include_contiguous,
    )

    connections: typing.Mapping[int, set[int]] = {
        r: set() for r in range(len(ranges))
    }
    for i, j in overlapping_ranges:
        connections[i].add(j)
        connections[j].add(i)

    group_of_each_range: typing.MutableMapping[int, set[int]] = {}
    for r in range(len(ranges)):

        # check which groups its neighbors are in
        groups_of_range = set()
        for other_r in connections[r]:
            if other_r in group_of_each_range:
                groups_of_range.add(other_r)

        group = {r}
        if len(groups_of_range) == 0:
            # if no neighbors are in groups, create a new one
            group |= set(connections[r])

        else:
            # if neighbors are in groups, combine those groups
            for other_r in groups_of_range:
                group |= group_of_each_range[other_r]

        for i in group:
            group_of_each_range[i] = group

    unique_groups = set(tuple(group) for group in group_of_each_range.values())

    combined_ranges = []
    for group_rs in unique_groups:
        range_start = min(ranges[group_range][0] for group_range in group_rs)
        range_end = max(ranges[group_range][1] for group_range in group_rs)
        combined_ranges.append([range_start, range_end])

    return combined_ranges


# def combine_overlapping_ranges(
#     ranges: typing.Sequence[Range],
#     *,
#     include_contiguous: bool = False,
# ) -> typing.Sequence[Range]:
#     """combine ranges as needed to produce list of non-overlapping ranges"""

#     print("RANGES (" + str(len(ranges)) + ')', ranges)
#     print()

#     overlapping_ranges = get_overlapping_ranges(
#         ranges,
#         include_contiguous=include_contiguous,
#     )

#     if len(overlapping_ranges) == 0:
#         return ranges

#     else:
#         combined_ranges = []
#         for i, j in overlapping_ranges:
#             combined_start = min(ranges[i][0], ranges[j][0])
#             combined_end = max(ranges[i][1], ranges[j][1])
#             combined_ranges.append([combined_start, combined_end])
#         fixed_ranges = combine_overlapping_ranges(
#             combined_ranges,
#             include_contiguous=include_contiguous,
#         )

#         overlapping_indices = {index for r in overlapping_ranges for index in r}
#         nonoverlapping_ranges = [
#             range
#             for r, range in enumerate(ranges)
#             if r not in overlapping_indices
#         ]

#         return nonoverlapping_ranges + list(fixed_ranges)


def range_to_chunks(
    *,
    start: int,
    end: int,
    chunk_size: int,
    round_bounds: bool = False,
    trim_outer_bounds: bool = False,
    index: bool = False,
) -> typing.Sequence[typing.Sequence[int]]:
    """break a range into chunks of a given chunk size

    ## Examples
    1. range_to_chunks(start=390, end=710, chunk_size=100)
        --> [[390, 489], [490, 589], [590, 689], [690, 710]]
    2. range_to_chunks(start=390, end=710, chunk_size=100, round_bounds=True)
        --> [[300, 399], [400, 499], [500, 599], [600, 699], [700, 799]
    3. range_to_chunks(start=390, end=710, chunk_size=100, round_bounds=True, trim_outer_bounds=True)
        --> [[390, 399], [400, 499], [500, 599], [600, 699], [700, 710]
    4. range_to_chunks(start=390, end=710, chunk_size=100, index=True)
        --> [[390, 490], [490, 590], [590, 690], [690, 711]]

    ## Inputs
    - start
    - end
    - chunk_size
    - round_bounds: round lower down and upper up to reach standardized bounds
    - trim_outer_bounds: trim standardized bounds to not exceed original inputs
        - has no effect unless round_bounds=True
    - index: return ranges that can be used as python indices (end = end + 1)
    """

    # compute range bounds
    if round_bounds:
        start_bound = (start // chunk_size) * chunk_size
        end_bound = ((end // chunk_size) + 1) * chunk_size
        bounds = list(range(start_bound, end_bound + 1, chunk_size))
    else:
        bounds = list(range(start, end + 1, chunk_size))
        bounds.append(end + 1)

    # create chunks
    chunks = []
    for bound_start, bound_end in zip(bounds[:-1], bounds[1:]):
        if index:
            chunk = [bound_start, bound_end]
        else:
            chunk = [bound_start, bound_end - 1]
        chunks.append(chunk)

    # trim outer bounds if needed
    if trim_outer_bounds:
        if len(chunks) > 0:

            if chunks[0][0] < start:
                chunks[0] = [start, chunks[0][1]]

            if index:
                last_value = end + 1
            else:
                last_value = end
            if chunks[-1][-1] > last_value:
                chunks[-1] = [chunks[-1][0], last_value]

    return chunks




from __future__ import annotations

import time
import typing

from .. import fourbyte_spec

if typing.TYPE_CHECKING:
    import asyncio
    from typing_extensions import TypedDict

    class FourbyteRatelimit(TypedDict):
        requests_per_second: int | float
        last_request_time: int | float
        lock: asyncio.Lock | None


_4byte_ratelimit: FourbyteRatelimit = {
    'lock': None,
    'last_request_time': 0,
    'requests_per_second': 4.0,
}


async def async_query_remote_function_signatures(
    hex_signature: typing.Optional[str] = None,
    *,
    id: typing.Optional[int] = None,
    bytes_signature: typing.Optional[str] = None,
    text_signature: typing.Optional[str] = None,
) -> typing.Sequence[fourbyte_spec.Entry]:
    import aiohttp

    # get url template
    if id is not None:
        url_template = fourbyte_spec.endpoints['function_id']
    elif hex_signature is not None:
        url_template = fourbyte_spec.endpoints['function_hex']
    elif text_signature is not None:
        url_template = fourbyte_spec.endpoints['function_text']
    else:
        raise Exception('must specify id, hex_signature, or text_signature')

    # get url
    inputs = {
        'id': id,
        'bytes_signature': bytes_signature,
        'hex_signature': hex_signature,
        'text_signature': text_signature,
    }
    url = url_template.format(**inputs)

    # acquire lock
    import asyncio

    lock = _4byte_ratelimit['lock']
    if lock is None:
        lock = asyncio.Lock()
        _4byte_ratelimit['lock'] = lock

    # perform request
    async with lock:
        # ratelimit
        now = time.time()
        time_since_last = now - _4byte_ratelimit['last_request_time']
        seconds_per_request = 1 / _4byte_ratelimit['requests_per_second']
        time_to_sleep = seconds_per_request - time_since_last
        if time_to_sleep > 0:
            if True:
                print(
                    '4byte ratelimit hit, sleeping for '
                    + str(time_to_sleep)
                    + ' seconds (build local 4byte db to avoid)'
                )
            await asyncio.sleep(time_to_sleep)

        # perform request
        async with aiohttp.ClientSession() as session:
            # acquire, and retry if initial bad response
            async with session.get(url) as response:
                if response.status == 502:
                    await asyncio.sleep(3)
                else:
                    result = await response.json()
            async with session.get(url) as response:
                result = await response.json()

            _4byte_ratelimit['last_request_time'] = time.time()

            # process result
            if id is not None:
                return [result]
            else:
                return result['results']  # type: ignore


async def async_query_remote_event_signatures(
    hex_signature: typing.Optional[str] = None,
    *,
    id: typing.Optional[int] = None,
    bytes_signature: typing.Optional[str] = None,
    text_signature: typing.Optional[str] = None,
) -> typing.Sequence[fourbyte_spec.Entry]:
    import aiohttp

    # get url template
    if id is not None:
        url_template = fourbyte_spec.endpoints['event_id']
    elif hex_signature is not None:
        url_template = fourbyte_spec.endpoints['event_hex']
    elif text_signature is not None:
        url_template = fourbyte_spec.endpoints['event_text']
    else:
        raise Exception('must specify id, hex_signature, or text_signature')

    # get url
    inputs = {
        'id': id,
        'bytes_signature': bytes_signature,
        'hex_signature': hex_signature,
        'text_signature': text_signature,
    }
    url = url_template.format(**inputs)

    # perform request
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            result = await response.json()
            if id is not None:
                return [result]
            else:
                return result['results']  # type: ignore




from __future__ import annotations

import os
import typing
import urllib.parse

import toolcli
import toolstr

from ctc import rpc
from ctc import spec
from ... import config_defaults


async def async_setup_networks(
    *,
    old_config: typing.Mapping[typing.Any, typing.Any],
    headless: bool,
    rpc_url: str | None,
    rpc_chain_id: int | None,
    skip_networking: bool = False,
    styles: toolcli.StyleTheme,
) -> spec.PartialConfig:

    if skip_networking:
        providers = old_config.get('providers')
        if providers is None:
            providers = {}

        networks = old_config.get('networks')
        if networks is None:
            networks = config_defaults.get_default_networks_metadata()

        default_network = old_config.get('default_network')
        if not isinstance(default_network, int):
            default_network = None

        default_providers = old_config.get('default_providers')
        if default_providers is None:
            default_providers = {}

        if default_network is None:
            if len(default_providers) > 0:
                default_network = min(default_providers.keys())
            else:
                default_network = 1
        return {
            'providers': providers,
            'networks': networks,
            'default_network': default_network,
            'default_providers': default_providers,
        }

    print()
    print()
    toolstr.print('## Network Setup', style=styles['title'])
    print()

    # get providers
    providers, networks = await async_specify_providers(
        old_config=old_config,
        styles=styles,
        headless=headless,
        rpc_url=rpc_url,
        rpc_chain_id=rpc_chain_id,
    )

    # get additional custom networks
    networks = specify_networks(
        networks=networks,
        styles=styles,
        headless=headless,
    )

    # get default network
    default_network = specify_default_network(
        providers=providers,
        networks=networks,
        styles=styles,
        headless=headless,
        rpc_url=rpc_url,
    )

    # get default providers
    default_providers = specify_default_providers(
        providers=providers,
        networks=networks,
        styles=styles,
        headless=headless,
        old_config=old_config,
    )

    print()
    print('Network setup complete')

    # return results
    data: spec.PartialConfig = {
        'providers': providers,
        'networks': networks,
        'default_network': default_network,
        'default_providers': default_providers,
    }
    return data


async def async_specify_providers(
    *,
    old_config: typing.Mapping[str, typing.Any],
    headless: bool,
    rpc_url: str | None,
    rpc_chain_id: int | None,
    styles: toolcli.StyleTheme,
) -> tuple[
    typing.Mapping[str, spec.Provider],
    typing.MutableMapping[spec.ChainId, spec.NetworkMetadata],
]:

    providers: typing.MutableMapping[str, spec.Provider] = {}
    networks: typing.MutableMapping[spec.ChainId, spec.NetworkMetadata] = dict(
        config_defaults.get_default_networks_metadata()
    )

    # add providers first, then configure their networks if those are unknown
    old_providers = old_config.get('providers', {})
    if len(old_providers) > 0:
        print('Currently using these providers:')
        for provider_name, provider_metadata in old_providers.items():
            toolstr.print(
                '-',
                provider_name,
                '['
                + styles['description']
                + ']'
                + provider_metadata['url']
                + '[/'
                + styles['description']
                + ']',
            )
        print()
        answer = toolcli.input_yes_or_no(
            'Would you like to continue using these providers? ',
            style=styles['metavar'],
            default='yes',
            headless=headless,
        )
        if answer:
            # TODO: validate old_providers
            providers.update(old_providers)

    if rpc_url is not None:
        await async_collect_provider_metadata(
            providers=providers,
            networks=networks,
            styles=styles,
            url=rpc_url,
            chain_id=rpc_chain_id,
            headless=headless,
        )
    elif len(providers) == 0:
        print()
        print('Most ctc operations require an RPC provider')
        print()
        prompt = 'Would you like to specify an RPC provider? '
        if toolcli.input_yes_or_no(
            prompt=prompt,
            style=styles['metavar'],
            default='yes',
            headless=headless,
        ):
            await async_collect_provider_metadata(
                providers=providers,
                networks=networks,
                styles=styles,
                headless=headless,
                url=None,
                chain_id=None,
            )

    # collect additional providers
    if len(providers) > 0:
        answer = toolcli.input_yes_or_no(
            prompt='Would you like to specify additional RPC providers? ',
            style=styles['metavar'],
            default='no',
            headless=headless,
        )
        while answer:

            # collect provider
            await async_collect_provider_metadata(
                providers=providers,
                networks=networks,
                styles=styles,
                headless=False,
                url=None,
                chain_id=None,
            )

            # prompt for additional providers
            answer = toolcli.input_yes_or_no(
                prompt='Would you like to specify additional RPC providers? ',
                style=styles['metavar'],
                default='no',
                headless=False,
            )

    return providers, networks


async def async_collect_provider_metadata(
    *,
    providers: typing.MutableMapping[str, spec.Provider],
    networks: typing.MutableMapping[spec.ChainId, spec.NetworkMetadata],
    styles: toolcli.StyleTheme,
    headless: bool,
    url: str | None,
    chain_id: int | None,
) -> None:
    """collect metadata for a provider"""

    print()
    if url is None:

        if os.environ.get('ETH_RPC_URL') not in [None, '']:
            default_url = os.environ['ETH_RPC_URL']
        else:
            default_url = None

        if headless and default_url is None:
            raise Exception(
                'if using headless mode, must either specify --rpc-url or set the ETH_RPC_URL env var'
            )

        url = toolcli.input_prompt(
            'What is the RPC provider URL? ',
            style=styles['metavar'],
            allow_blank=False,
            default=default_url,
            headless=headless,
        )
    else:
        print('Adding RPC provider: ' + str(url))

    if url.startswith('ws://'):
        # TODO: take new input instead of raising exception
        raise NotImplementedError('websockets not supported')
    if not (url.startswith('http://') or url.startswith('https://')):
        check_provider = rpc.resolve_provider(url)
        if check_provider['url'] != url:
            print()
            print('added prefix to url: ' + str(check_provider['url']))
            url = check_provider['url']

    if (
        chain_id is None
        and url == os.environ.get('ETH_RPC_URL')
        and os.environ.get('ETH_RPC_CHAIN_ID') not in [None, '']
    ):
        try:
            chain_id = int(os.environ['ETH_RPC_CHAIN_ID'])
            print(
                'could not use value of ETH_RPC_CHAIN_ID, not a valid integer'
            )
        except Exception:
            pass

    if chain_id is None:
        try:
            try:
                chain_id = rpc.rpc_provider._sync_get_chain_id(url)
            except Exception:
                chain_id = toolcli.input_int(
                    'Could not connect to RPC provider. What is this provider\'s chain_id? ',
                    style=styles['metavar'],
                    headless=False,
                )
            description = 'chain_id = ' + str(chain_id)
            if chain_id in networks:
                name = networks[chain_id]['name']
                if name is not None:
                    description = description + ', network = ' + name
            print('Provider using: ' + description)
        except Exception as e:
            raise e
            print('Could not query node for chain_id metadata')
            chain_id = toolcli.input_int(
                'What is the chain_id used by this node? ',
                style=styles['metavar'],
                headless=False,
            )

    # determine whether chain_id is of known network
    known_network = any(
        chain_id == network.get('chain_id') for network in networks.values()
    )

    # if chain_id of unknown network, collect network metadata
    if not known_network:
        if headless:
            raise NotImplementedError(
                'cannot use unknown network in headless mode'
            )
        collect_network_metadata(
            chain_id=chain_id,
            networks=networks,
            styles=styles,
        )

    name = toolcli.input_prompt(
        prompt='What should this node be called? ',
        default=create_default_provider_name(url=url, network=chain_id),
        style=styles['metavar'],
        headless=headless,
    )
    if url.startswith('http'):
        protocol: typing.Literal['http'] = 'http'
    else:
        raise Exception('unknown protocol, missing http(s) in url?')
    provider: spec.Provider = {
        'name': name,
        'url': url,
        'network': chain_id,
        'protocol': protocol,
        'session_kwargs': {},
        'chunk_size': None,
        'convert_reverts_to_none': False,
        'disable_batch_requests': False,
    }
    providers[name] = provider


def create_default_provider_name(url: str, network: int) -> str:
    if all(c.isnumeric() or c in ['.', ':'] for c in url.split('://')[1]):
        return url.split('://')[1].split('/')[0] + '__' + str(network)

    hostname = urllib.parse.urlparse(url).hostname
    if hostname is not None:
        hostname_pieces = hostname.split('.')
        if len(hostname_pieces) == 1:
            hostname_piece = hostname_pieces[0]
        else:
            hostname_piece = hostname_pieces[-2]
    else:
        hostname_piece = url

    return hostname_piece + '__' + str(network)


def collect_network_metadata(
    *,
    styles: toolcli.StyleTheme,
    networks: typing.MutableMapping[spec.ChainId, spec.NetworkMetadata],
    name: str | None = None,
    chain_id: int | None = None,
) -> None:
    """collect metadata for a network"""

    if chain_id is None:
        chain_id = toolcli.input_int(
            'What is the network\'s chain_id? (enter a blank line to go back)\n',
            style=styles['metavar'],
        )
        # CHECK that chain_id is not already taken
    if name is None:
        name = toolcli.input_prompt(
            'What is the network\'s name? (enter a blank line to go back)\n',
            style=styles['metavar'],
        )
        # CHECK that name is not already taken

    block_explorer = toolcli.input_prompt(
        'Network block explorer? ', style=styles['metavar']
    )
    network_metadata: spec.NetworkMetadata = {
        'name': name,
        'chain_id': chain_id,
        'block_explorer': block_explorer,
    }
    networks[chain_id] = network_metadata


def specify_networks(
    *,
    networks: typing.MutableMapping[spec.ChainId, spec.NetworkMetadata],
    styles: toolcli.StyleTheme,
    headless: bool,
) -> typing.MutableMapping[spec.ChainId, spec.NetworkMetadata]:

    # print current networks
    print()
    print('Have metadata for the following networks:')
    print()
    rows = [
        [networks[chain_id]['name'], chain_id]
        for chain_id in sorted(networks.keys())
    ]
    toolstr.print_table(
        rows,
        labels=['name', 'chain_id'],
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'name': styles['metavar'],
            'chain_id': styles['description'],
        },
        add_row_index=True,
    )

    # add new networks
    while toolcli.input_yes_or_no(
        '\nWould you like to add metadata for additional networks? ',
        style=styles['metavar'],
        default='no',
        headless=headless,
    ):
        collect_network_metadata(
            styles=styles,
            networks=networks,
        )

    print()
    print('Using', len(networks), 'networks in config')

    return networks


def specify_default_network(
    *,
    networks: typing.Mapping[spec.ChainId, spec.NetworkMetadata],
    providers: typing.Mapping[str, spec.Provider],
    styles: toolcli.StyleTheme,
    headless: bool,
    rpc_url: str | None,
) -> spec.ChainId:

    # set default network
    choices_set = [
        str(network['name']) + ' (chain_id = ' + str(network['chain_id']) + ')'
        for network in networks.values()
    ]
    choices = sorted(choices_set)

    # only consider given provider if rpc_url is present
    if rpc_url is not None:
        providers = {
            provider_name: provider
            for provider_name, provider in providers.items()
            if provider['url'] == rpc_url
        }

    # determine default choice
    default = 'ethereum (chain_id = 1)'
    if len(providers) == 1:
        provider = list(providers.values())[0]
        network = provider.get('network')
        for network_metadata in networks.values():
            if (
                isinstance(network, int)
                and network == network_metadata['chain_id']
            ) or (
                isinstance(network, str) and network == network_metadata['name']
            ):
                default = (
                    str(network_metadata['name'])
                    + ' (chain_id = '
                    + str(network_metadata['chain_id'])
                    + ')'
                )
                break
    elif len(providers) > 1:
        for provider in providers.values():
            if provider.get('network') in [1, 'ethereum']:
                default = 'ethereum (chain_id = 1)'

    print()
    default_network_index = toolcli.input_number_choice(
        prompt='Which network to use as default?',
        choices=choices,
        default=default,
        style=styles['metavar'],
        headless=headless,
    )

    # clean this horrorshow up
    return int(choices[default_network_index].split(' ')[-1].strip(')'))


def specify_default_providers(
    *,
    networks: typing.Mapping[spec.ChainId, spec.NetworkMetadata],
    providers: typing.Mapping[str, spec.Provider],
    styles: toolcli.StyleTheme,
    headless: bool,
    old_config: typing.Mapping[typing.Any, typing.Any],
) -> typing.Mapping[spec.ChainId, str]:

    # compile providers for each network
    providers_per_network: dict[int, list[str]] = {}
    for provider_name, provider_metadata in providers.items():
        network = provider_metadata['network']
        if network is None:
            raise Exception(
                'unknown network for provider: ' + str(provider_metadata)
            )
        if not isinstance(network, int):
            for chain_id, network_metadata in networks.items():
                if network_metadata['name'] == network:
                    network = chain_id
                    break
            else:
                raise Exception(
                    'could not determine chain_id for network: ' + str(network)
                )
        providers_per_network.setdefault(network, [])
        providers_per_network[network].append(provider_name)

    # get default provider for each network
    default_providers = {}
    for network in providers_per_network:
        n_providers = len(providers_per_network[network])
        if n_providers == 1:
            default_providers[network] = providers_per_network[network][0]
        elif n_providers > 1:
            prompt = (
                'Which provider to use as default for '
                + str(networks[network]['name'])
                + ', chain_id = '
                + str(network)
                + '?'
            )

            default_choice = old_config.get('default_providers', {}).get(
                str(network)
            )
            if (
                default_choice is None
                or default_choice not in providers_per_network[network]
            ):
                default_choice = providers_per_network[network][0]

            answer = toolcli.input_number_choice(
                prompt=prompt,
                choices=providers_per_network[network],
                style=styles['metavar'],
                headless=headless,
                default=default_choice,
            )
            default_providers[network] = providers_per_network[network][answer]

    return default_providers




from __future__ import annotations

import toolcli
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': charset_command,
        'help': 'change character sets used for plotting',
        'args': [
            {
                'name': 'charset',
                'help': 'charset to switch to',
                'nargs': '?',
            },
            {
                'name': ['-v', '--verbose'],
                'help': 'display available charsets',
                'action': 'store_true',
            },
        ],
        'examples': [
            '',
            'braille',
            'quadrants',
            '-v',
        ]
    }


def charset_command(charset: str | None, verbose: bool) -> None:

    if verbose:
        cli_utils.preview_available_charsets()
    elif charset is None:
        cli_utils.print_current_charset()
    else:
        cli_utils.set_charset(charset)
